"""Utilities for working with the file system."""
import os
import time
import pathlib
from typing import Optional
import urllib.error
import urllib.request
import zipfile

from absl import logging

import inductiva


def get_timestamped_path(path_str: str, sep: str = "-") -> pathlib.Path:
    """Return a path that does not exist by appending a timestamp.

    Args:
        path: Path to a file or directory.

    Returns:
        A path that does not exist by appending the timestamp.
    """
    path = pathlib.Path(path_str)
    timestamp = time.strftime("%Y-%m-%dT%Hh%Mm%Ss")

    name = f"{path.stem}{sep}{timestamp}"

    return path.with_name(name + path.suffix)


def get_path_size(path_str: str) -> float:
    """Return the size of a path in bytes.
    
    Args:
        path: Path to a file or directory.
    
    Returns:
        The size of the path in bytes.
    """

    path = pathlib.Path(path_str)

    if not path.exists():
        raise FileNotFoundError(f"Path '{path}' does not exist.")

    if path.is_file():
        return path.stat().st_size

    size = 0

    for root, _, files in os.walk(path):
        root = pathlib.Path(root)
        size += root.stat().st_size
        for file in files:
            fp = root / file
            size += fp.stat().st_size
    return size


def resolve_output_path(path: Optional[str]) -> pathlib.Path:
    """Resolve a path relative to the output_dir

    Args:
        path: Path to a file or directory.
    """
    resolved_path = pathlib.Path.cwd()

    if inductiva.get_output_dir():
        resolved_path = pathlib.Path(inductiva.get_output_dir())

    if path is not None:
        resolved_path = pathlib.Path(resolved_path, path)

    return resolved_path


def get_sorted_files(data_dir: str,
                     file_format: str = "name",
                     split_token: str = "_"):
    """Returns list of files sorted according to [file_key].

    Order a set of .format files of the form
    ['name_1.format', 'name_2.format',...,'name_10.format',
    ...,'name_n.format'].

    The default sorting methods for list, list.sort()
    or sorted(list), order 'name_10.format' before 'name_2.format',
    which is not representative of the time series.

    In this function we sort according to the number..
    """

    if not os.path.exists(data_dir):
        raise IOError(f"Directory '{data_dir}' does not exist.")

    # Get a list of the files in the data directory.
    files = os.scandir(data_dir)

    # The files have file_format extension.
    files = [
        file for file in files if pathlib.Path(file.path).suffix == file_format
    ]

    # Sort the files to be read according to [file_key].
    def get_alphanum_key(file):
        file_name = pathlib.Path(file.path).stem
        file_name_splits = file_name.split(split_token)
        file_key = file_name_splits[-1]
        return int(file_key)

    files = sorted(files, key=get_alphanum_key)

    return files


def _unzip(zip_path: pathlib.Path):
    """Unzip a zip archive and remove the .zip."""

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall()

    zip_path.unlink()


def download_from_url(url: str, unzip: bool = False) -> str:
    """Download a file from an URL.

    This function downloads from a set of files from an url.
    If the file is zipped, then the argument unzip allows to
    unzip everything exactly as it was zipped before.

    Args:
        url: The URL to download the file from.
        unzip: Whether to unzip the file after downloading.
    Returns:
        The path to the downloaded file.
    """
    # Get archive name from url passed
    local_path = url.split("/")[-1]
    local_path = pathlib.Path(local_path)

    try:
        logging.info("Downloading from URL to the local path: %s", local_path)
        downloaded_to, _ = urllib.request.urlretrieve(url, filename=local_path)
    except urllib.error.URLError as url_error:
        logging.error("Could not download file from %s", url)
        raise url_error

    # Unzip all files as they were zipped.
    if unzip and zipfile.is_zipfile(downloaded_to):
        local_path = local_path.with_suffix("")
        logging.info("■ Uncompressing the downloaded file to: %s", local_path)
        _unzip(downloaded_to)
        logging.info("")
    return str(local_path.absolute())



"""Simsopt module of the API."""
from typing import Optional

from inductiva import simulators, tasks, types


class SIMSOPT(simulators.Simulator):
    """Invokes a simsopt simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the SIMOPT simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.api_method_name = "stellarators.simsopt.run_simulation"

    def _get_image_uri(self):
        return None

    def run(
        self,
        input_dir: str,
        plasma_surface_filename: str,
        coil_coefficients_filename: str,
        coil_currents_filename: str,
        num_field_periods: int,
        num_iterations: int,
        num_samples: int,
        sigma_scaling_factor: float,
        objectives_weights_filename: str,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        extra_metadata: Optional[dict] = None,
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            coil_coefficients_filename: Name of the file with the Fourier
              Series coefficients of the coils.
            coil_currents_filename: Name of the file with the current in each
              coils.
            plasma_surface_filename: Name of the file with the description of
              the plasma surface on which the magnetic field will be calculated.
            num_field_periods: Number of magnetic field periods.
              Refers to the number of complete magnetic field repetitions
              within a stellarator. Represents how many times the magnetic
              field pattern repeats itself along the toroidal direction.
            num_iterations: Number of iterations to run for the searching
              process.
            num_samples: Number of different stellarator samples generated per
              iteration from a configuration. The samples are generated using
              normal distribution noise for each coefficient of the Fourier
              Series that describes the coils.
            sigma_scaling_factor: Scaling factor for the sigma value used in
              the random generation of noise. This argument makes sure that
              the noise is generated proportionally to each coefficient. It
              also determines the range of search for new values of the
              coefficients.
            objectives_weights_filename: Name of the file with the weights for
              each objective function used in the construction of the total
              objective.
            storage_dir: Directory for storing results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiates with
                `spot=True`.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            objectives_weights_filename=objectives_weights_filename,
            coil_coefficients_filename=coil_coefficients_filename,
            plasma_surface_filename=plasma_surface_filename,
            coil_currents_filename=coil_currents_filename,
            resubmit_on_preemption=resubmit_on_preemption,
            sigma_scaling_factor=sigma_scaling_factor,
            num_field_periods=num_field_periods,
            num_iterations=num_iterations,
            extra_metadata=extra_metadata,
            num_samples=num_samples,
            storage_dir=storage_dir,
            on=on,
            **kwargs,
        )



"""Remove the user's remote storage contents via CLI."""
import argparse
import sys

from inductiva import storage
from inductiva.client import exceptions
from inductiva.utils.input_functions import user_confirmation_prompt
from ...localization import translator as __


def remove(args):
    """Remove user's remote storage contents."""
    paths = args.path
    confirm = args.confirm
    all_paths = args.all

    if not all_paths and not paths:
        print("No path(s) specified.\n"
              "> Use `inductiva storage remove -h` for help.")
        return 1

    if paths and all_paths:
        print(
            "inductiva storage remove: error: "
            "argument path not allowed with argument --all",
            file=sys.stderr)
        return 1
    paths = set(paths)
    if not confirm:
        confirm = user_confirmation_prompt(
            paths, __("storage-prompt-remove-all"),
            __("storage-prompt-remove-big", len(paths)),
            __("storage-prompt-remove-small"), all_paths)

    if confirm:
        if all_paths:
            storage.rmdir("*", confirm=True)
        for path in paths:
            try:
                storage.rmdir(path, confirm=True)
            except exceptions.ApiValueError as rmdir_exception:
                print(rmdir_exception)

    return 0


def register(parser):
    subparser = parser.add_parser("remove",
                                  aliases=["rm"],
                                  help="Remove remote storage entries.",
                                  formatter_class=argparse.RawTextHelpFormatter)
    subparser.description = (
        "The `inductiva storage remove` command deletes specified data"
        " from the platform.\n"
        "It targets a specific path for removal. Use with caution as "
        "this action is irreversible.\n\n"
        "Use `inductiva storage remove \"*\"` to remove all folders in"
        " your storage.")
    subparser.add_argument("path",
                           type=str,
                           nargs="*",
                           help="Path(s) to be removed from remote storage. "
                           "To remove all contents, use \"*\".")

    subparser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        dest="confirm",
        default=False,
        help="Sets any confirmation values to \"yes\" "
        "automatically. Users will not be asked for "
        "confirmation to remove path(s) from remote storage.")
    subparser.add_argument("--all",
                           action="store_true",
                           default=False,
                           help="Remove all contents from remote storage.")

    subparser.set_defaults(func=remove)



#pylint: disable=missing-module-docstring
from .machines import MachineGroup, ElasticMachineGroup
from . import machines_base
from . import machine_groups
from .machine_groups import estimate_machine_cost
from .machine_cluster import MPICluster
from .machine_types import list_available_machines
from .machine_groups import get



"""CLI utils."""
import argparse
import pathlib
import shutil
import os
import re

import inductiva
from inductiva import logs

BEGIN_TAG = "# >>> INDUCTIVA BEGIN:"
END_TAG = "# >>> INDUCTIVA END:"


def positive_float(value):
    """Check if the given value is a positive float."""
    try:
        value = float(value)
    except TypeError as _:
        raise argparse.ArgumentTypeError(f"{value} is an invalid type.")
    if value <= 0:
        raise argparse.ArgumentTypeError(f"{value} must be a positive number.")
    return value


def add_watch_argument(subparser):
    """Add watch flag to the subparser that prompts the commands every Nsecs."""
    subparser.set_defaults(watchable=True)
    subparser.add_argument("-w",
                           "--watch",
                           nargs="?",
                           const=2.0,
                           type=positive_float,
                           help="Prompt the command every N seconds.")


def check_running_for_first_time():
    """Checks if it is the first time the cli is being called.

    It does this by checking if there is a folder:

      `$HOME/.inductiva/v{inductiva_version}`

    If there is not returns true and creates that folder.

    """
    version = inductiva.__version__
    dir_name = logs.log.get_logs_file_path().parent / f"v{version}"
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)
        return True
    return False


def user_autocompletion_install_prompt():
    prompt = input("This is the first time running the inductiva cli. "
                   "Would you like to enable autocompletion? (y/[N])?\n"
                   "At the moment only zsh is supported.")
    return prompt.lower() in ["y", "ye", "yes"]


def show_help_msg(parser):
    """Show help message for command if no subcommand is provided."""
    parser.set_defaults(func=lambda _: parser.print_help())


def setup_zsh_autocompletion():
    """Sets up shell auto-completion for zsh.

    It does this by:

    1. Adding the file with completions to:
      `$HOME/.inductiva/v{version}/completions/_inductiva`
    This file already comes with the `inductiva` package.

    2. Adding to the users .zshrc:

        # >>> INDUCTIVA BEGIN:
        # !! Contents within this block are managed by 'inductiva' !!
        # !! Removal might break some functionalities !!
        source {setup_file_path}
        # >>> INDUCTIVA END:

    That points to a file:
    /.inductiva/v{version}/inductiva-setup.sh. This file than adds the
    autocompletion script to the `fpath` of the user.

    """
    version = inductiva.__version__
    version_dir = logs.log.get_logs_file_path().parent / f"v{version}"
    assets_dir = pathlib.Path(inductiva.__path__[0]) / "assets"
    shutil.copytree(assets_dir, version_dir, dirs_exist_ok=True)

    lines_to_append = [
        f"\n\n{BEGIN_TAG}\n",
        "# !! Contents within this block are managed by 'inductiva' !!\n",
        "# !! Removal might break some functionalities !!\n",
        f"source {version_dir / 'inductiva-setup.sh'}\n"
        f"{END_TAG}\n\n",
    ]

    regex = re.compile(f"{BEGIN_TAG}.*{END_TAG}", re.DOTALL)
    zshrc_path = pathlib.Path.home() / ".zshrc"

    with open(zshrc_path, "r", encoding="utf8") as f:
        content = f.read()

    with open(zshrc_path, "w", encoding="utf8") as f:
        content = regex.sub("", content)
        f.seek(0)
        f.truncate()
        lines_as_string = "".join(lines_to_append)
        content += f"\n{lines_as_string}\n"
        f.write(content)



# pylint: disable=missing-module-docstring
from .simulator import Simulator
from .dummy_simulator import DummySimulator
from .splishsplash import SplishSplash
from .dualsphysics import DualSPHysics
from .openfoam import OpenFOAM
from .openfast import OpenFAST
from .amr_wind import AmrWind
from .gromacs import GROMACS
from .simsopt import SIMSOPT
from .fenicsx import FEniCSx
from .reef3d import REEF3D
from .xbeach import XBeach
from .schism import SCHISM
from .nwchem import NWChem
from .swash import SWASH
from .cans import CaNS
from .swan import SWAN
from .fds import FDS



from inductiva.client.paths.tasks_task_id.get import ApiForget
from inductiva.client.paths.tasks_task_id.delete import ApiFordelete


class TasksTaskId(
        ApiForget,
        ApiFordelete,
):
    pass



"""Get the user's remote storage size in GB via CLI."""

import argparse
from inductiva import storage


def storage_used(unused_args):
    """List the user's remote storage contents."""
    storage.get_space_used()


def register(parser):
    """Register the list user's remote storage command."""

    subparser = parser.add_parser("size",
                                  help="Get the occupied remote storage size.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva storage size` command calculates "
        "the total size of your data on the platform.\n"
        "It returns the total size in GB of all items in your storage.\n")

    subparser.set_defaults(func=storage_used)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ValidationError(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "msg",
            "loc",
            "type",
        }

        class properties:

            class loc(schemas.ListSchema):

                class MetaOapg:

                    class items(
                            schemas.ComposedSchema,):

                        class MetaOapg:
                            any_of_0 = schemas.StrSchema
                            any_of_1 = schemas.IntSchema

                            @classmethod
                            @functools.lru_cache()
                            def any_of(cls):
                                # we need this here to make our import statements work
                                # we must store _composed_schemas in here so the code is only run
                                # when we invoke this method. If we kept this at the class
                                # level we would get an error because the class level
                                # code would be run when this module is imported, and these composed
                                # classes don't exist yet because their module has not finished
                                # loading
                                return [
                                    cls.any_of_0,
                                    cls.any_of_1,
                                ]

                        def __new__(
                            cls,
                            *_args: typing.Union[
                                dict,
                                frozendict.frozendict,
                                str,
                                date,
                                datetime,
                                uuid.UUID,
                                int,
                                float,
                                decimal.Decimal,
                                bool,
                                None,
                                list,
                                tuple,
                                bytes,
                                io.FileIO,
                                io.BufferedReader,
                            ],
                            _configuration: typing.Optional[
                                schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                                   frozendict.frozendict, str,
                                                   date, datetime, uuid.UUID,
                                                   int, float, decimal.Decimal,
                                                   None, list, tuple, bytes],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[
                        MetaOapg.items,
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ]], typing.List[typing.Union[
                        MetaOapg.items,
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ]]],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'loc':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)

            msg = schemas.StrSchema
            type = schemas.StrSchema
            __annotations__ = {
                "loc": loc,
                "msg": msg,
                "type": type,
            }

    msg: MetaOapg.properties.msg
    loc: MetaOapg.properties.loc
    type: MetaOapg.properties.type

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["loc"]) -> MetaOapg.properties.loc:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["msg"]) -> MetaOapg.properties.msg:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["type"]
    ) -> MetaOapg.properties.type:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "loc",
        "msg",
        "type",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["loc"]) -> MetaOapg.properties.loc:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["msg"]) -> MetaOapg.properties.msg:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["type"]
    ) -> MetaOapg.properties.type:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "loc",
        "msg",
        "type",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        msg: typing.Union[
            MetaOapg.properties.msg,
            str,
        ],
        loc: typing.Union[
            MetaOapg.properties.loc,
            list,
            tuple,
        ],
        type: typing.Union[
            MetaOapg.properties.type,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'ValidationError':
        return super().__new__(
            cls,
            *_args,
            msg=msg,
            loc=loc,
            type=type,
            _configuration=_configuration,
            **kwargs,
        )



"""Manage running/completed tasks on the Inductiva API."""
import pathlib
import contextlib
import sys
import time
import json
from absl import logging
from typing import Dict, Any, List, Optional, Tuple, Union
from typing_extensions import TypedDict
import datetime
from ..localization import translator as __
import urllib3
import tabulate
from dataclasses import dataclass

from inductiva import constants
from inductiva.client import exceptions, models
from inductiva import api
from inductiva.client.apis.tags import tasks_api
from inductiva.utils import files, format_utils, data
from inductiva.tasks import output_info

import warnings


@dataclass
class Metric:
    """Represents a single metric with a value and a label."""
    label: str
    value: Optional[float] = None


class TaskInfo:
    """Represents the task information."""

    MISSING_UNTIL_TASK_STARTED = "N/A until task is started"
    MISSING_UNTIL_TASK_ENDED = "N/A until task ends"

    class Executer:
        """Encapsulates information about the executer."""

        def __init__(self):
            self.uuid = None
            self.cpu_count_logical = None
            self.cpu_count_physical = None
            self.memory = None
            self.n_mpi_hosts = None
            self.vm_type = None
            self.vm_name = None
            self.host_type = None
            self.error_detail = None

    class TimeMetrics:
        """Encapsulates time-related metrics."""

        def __init__(self):
            self.total_seconds = Metric("Total seconds")
            self.input_upload_seconds = Metric("Input upload")
            self.queue_time_seconds = Metric("Time in queue")
            self.container_image_download_seconds = Metric(
                "Container image download")
            self.input_download_seconds = Metric("Input download")
            self.input_decompression_seconds = Metric("Input decompression")
            self.computation_seconds = Metric("Computation")
            self.output_compression_seconds = Metric("Output compression")
            self.output_upload_seconds = Metric("Output upload")

    class DataMetrics:
        """Encapsulates data-related metrics."""

        def __init__(self):
            self.output_zipped_size_bytes = Metric("Size of zipped output")
            self.output_size_bytes = Metric("Size of unzipped output")
            self.output_total_files = Metric("Number of output files")

    def __init__(self, **kwargs):
        """Initialize the instance with attributes from keyword arguments."""
        self.is_submitted = None
        self.is_running = None
        self.is_terminal = None
        self.task_id = None
        self.status = None
        self.method_name = None
        self.storage_path = None
        self.container_image = None
        self.project = None
        self.create_time = None
        self.input_submit_time = None
        self.start_time = None
        self.computation_start_time = None
        self.computation_end_time = None
        self.end_time = None
        self.time_metrics = self.TimeMetrics()
        self.data_metrics = self.DataMetrics()

        # Set the general attributes
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

        # Set the metrics objects
        metrics_info: dict = kwargs.get("metrics", {})
        self.__update_metrics(self.time_metrics, metrics_info)
        self.__update_metrics(self.data_metrics, metrics_info)

        # Set the executer object
        self.executer = None
        executer_info: dict = kwargs.get("executer", {})
        if executer_info:
            self.executer = self.Executer()
            for key, value in executer_info.items():
                if hasattr(self.executer, key):
                    setattr(self.executer, key, value)

        # Update running info
        self.is_submitted = self.status == models.TaskStatusCode.SUBMITTED
        self.is_running = self.status == models.TaskStatusCode.STARTED
        self.is_terminal = kwargs.get("is_terminated", False)

    def __update_metrics(
        self,
        metrics_obj: Metric,
        metrics_info: Dict[str, Any],
    ) -> None:
        for key, value in metrics_info.items():
            if key in metrics_obj.__dict__:
                getattr(metrics_obj, key).value = value

    def _format_time_metric(
        self,
        metric_key: str,
        value: Optional[float],
    ) -> str:
        if isinstance(value, float):
            if value >= 60.0:
                value_str = format_utils.seconds_formatter(value)
            else:
                value_str = f"{value:.2f} s"

            if metric_key == "computation_seconds" and self.is_running:
                value_str += " (Task still running)"

            return value_str

        # Value is None if it is not float
        if self.is_terminal:
            if metric_key == "container_image_download_seconds":
                # If the container image is already present in the local cache
                # the download is skipped, therefore the metric does not exist
                return "N/A (used cached image)"
            if metric_key == "computation_seconds":
                # The task might have ended but the metric is not available in
                # the database yet
                return "N/A"

        if metric_key == "container_image_download_seconds":
            # If the task has not ended the local cache image may be used or
            # the download could be in progress
            return "N/A"

        if metric_key in (
                "output_compression_seconds",
                "output_upload_seconds",
        ):
            return self.MISSING_UNTIL_TASK_ENDED

        # None values will be replaced with a default missing value message
        return value

    @staticmethod
    def _format_data_metric(
        metric_key: str,
        value: Optional[float],
    ) -> Optional[str]:
        if isinstance(value, int) and "bytes" in metric_key:
            return format_utils.bytes_formatter(value)
        return value

    def __repr__(self) -> str:
        return str(self)

    def __str__(self):
        table_format = "plain"

        wall_time_data = [[
            "Wall clock time:",
            self._format_time_metric(
                "total_seconds",
                self.time_metrics.total_seconds.value,
            ),
        ]]
        wall_time_table = tabulate.tabulate(
            wall_time_data,
            tablefmt=table_format,
        )

        time_metrics_data = [
            [
                f"{metric.label}:",
                self._format_time_metric(metric_key, metric.value),
            ]
            for metric_key, metric in self.time_metrics.__dict__.items()
            if metric_key != "total_seconds"
        ]
        time_metrics_table = tabulate.tabulate(
            time_metrics_data,
            missingval=self.MISSING_UNTIL_TASK_STARTED,
            tablefmt=table_format,
        )

        data_metrics_data = [[
            f"{metric.label}:",
            self._format_data_metric(metric_key, metric.value)
        ] for metric_key, metric in self.data_metrics.__dict__.items()]
        data_metrics_table = tabulate.tabulate(
            data_metrics_data,
            missingval=self.MISSING_UNTIL_TASK_ENDED,
            tablefmt=table_format,
        )

        # Add a tab to the beginning of each line in the tables
        time_metrics_table = "\n".join(
            "\t" + line for line in time_metrics_table.splitlines())
        data_metrics_table = "\n".join(
            "\t" + line for line in data_metrics_table.splitlines())

        table_str = f"\nTask status: {self.status}"
        if self.executer and self.executer.error_detail:
            table_str += f"\n\tStatus detail: {self.executer.error_detail}"
        table_str += f"\n{wall_time_table}"
        table_str += f"\nTime breakdown:\n{time_metrics_table}"
        table_str += f"\nData:\n{data_metrics_table}\n"

        return table_str


class Task:
    """Represents a running/completed task on the Inductiva API.

    Example usage:
        task = scenario.simulate(...)
        final_status = task.wait()
        info = task.get_info() # dictionary with info about the task
        task.download_outputs(
            filenames=["file1.txt", "file2.dat"] # download only these files
        )

    Attributes:
        id: The task ID.
        _api: Instance of TasksApi (from the generated API client).
    """

    FAILED_STATUSES = {
        models.TaskStatusCode.FAILED,
        models.TaskStatusCode.KILLED,
        models.TaskStatusCode.EXECUTERFAILED,
        models.TaskStatusCode.EXECUTERTERMINATED,
        models.TaskStatusCode.EXECUTERTERMINATEDBYUSER,
        models.TaskStatusCode.SPOTINSTANCEPREEMPTED,
        models.TaskStatusCode.ZOMBIE,
        models.TaskStatusCode.EXECUTERTERMINATEDTTLEXCEEDED,
        models.TaskStatusCode.TTLEXCEEDED,
    }

    RUNNING_STATUSES = {
        models.TaskStatusCode.PENDINGINPUT, models.TaskStatusCode.STARTED
    }

    KILLABLE_STATUSES = {models.TaskStatusCode.SUBMITTED
                        }.union(RUNNING_STATUSES)

    KILL_VERBOSITY_LEVELS = [0, 1, 2]

    STANDARD_OUTPUT_FILES = ["stdout.txt", "stderr.txt"]

    def __init__(self, task_id: str):
        """Initialize the instance from a task ID."""
        self.id = task_id
        self._api = tasks_api.TasksApi(api.get_client())
        self._info = None
        self._status = None
        self._tasks_ahead: Optional[int] = None
        self._summary = None
        # Internal state to track if the method was called from the wait method
        self._called_from_wait = False

    def is_running(self) -> bool:
        """Validate if the task is running.

        This method issues a request to the API.
        """
        return self.get_status() == models.TaskStatusCode.STARTED

    def is_failed(self) -> bool:
        """Validate if the task has failed.

        This method issues a request to the API.
        """
        return self.get_status() in self.FAILED_STATUSES

    def is_terminal(self) -> bool:
        """Check if the task is in a terminal status.

        This method issues a request to the API.
        """
        self.get_info()
        return self.info.is_terminal

    @classmethod
    def from_api_info(cls, info: Dict[str, Any]) -> "Task":

        task = cls(info["task_id"])
        task._info = TaskInfo(**info)
        task._status = models.TaskStatusCode(info["status"])

        # TODO(luispcunha): construct correct output class from API info.

        return task

    @contextlib.contextmanager
    def sync_context(self):
        """Enter context manager for blocking sync execution.

        This turns an asynchronous task into a blocking task.
        If an exception/ctrl+c is caught while in the context manager, the
        remote task is killed.

        Usage:
            task = scenario.simulate(...)

            with task.sync_context():
                # If an exception happens here or ctrl+c is pressed, the task
                # will be killed.
                task.wait()

        """
        try:
            yield
        except KeyboardInterrupt:
            logging.info("Caught SIGINT: terminating blocking task...")
            self.kill()
        except Exception as e:
            logging.error("Caught exception: terminating blocking task...")
            self.kill()
            raise e

    def get_status(self) -> models.TaskStatusCode:
        """Get status of the task.

        This method issues a request to the API and updates the task info
        is_terminal. The api call to get the status now returns the status and
        is_terminated.
        """
        # If the task is in a terminal status and we already have the status,
        # return it without refreshing it from the API.
        if (self._status is not None and self._info.is_terminal):
            return self._status

        resp = self._api.get_task_status(self._get_path_params())

        #updates the info.is_terminal when getting the status
        self.info.is_terminal = resp.body.get("is_terminated",
                                              self.info.is_terminal)

        queue_position = resp.body.get("position_in_queue", None)
        if queue_position is not None:
            self._tasks_ahead = queue_position.get("tasks_ahead", None)

        return models.TaskStatusCode(resp.body["status"])

    def get_position_in_queue(self) -> Optional[int]:
        """Get the position of the task in the queue.

        This method issues a request to the API.
        """
        try:
            resp = self._api.get_task_position_in_queue(self._get_path_params())
            self._tasks_ahead = resp.body.get("tasks_ahead", None)
            return self._tasks_ahead
        except exceptions.ApiException as exc:
            if exc.status == 404:
                return None

    @property
    def info(self) -> TaskInfo:
        """Get information about the task.

        It contains cached information about the task from the latest call to
        `get_info`, therefore it can be outdated.
        """
        if self._info is None:
            return self.get_info()
        return self._info

    def get_info(self) -> TaskInfo:
        """Get a dictionary with information about the task.

        Information includes e.g., "task_id", "status", timestamps
        ("create_time", "input_submit_time, "start_time", "end_time"),
        among others.

        This method issues a request to the API.
        """
        params = self._get_path_params()
        resp = self._api.get_task(params, skip_deserialization=True).response

        info = json.loads(resp.data.decode("utf-8"))
        status = models.TaskStatusCode(info["status"])

        self._info = TaskInfo(**info)
        self._status = status

        return self._info

    def _setup_queue_message(self, is_tty: bool) -> str:
        if self._tasks_ahead == 0:
            s = f"The task {self.id} is about to start."
        else:
            s = (f"Number of tasks ahead of task {self.id} in queue: "
                 f"{self._tasks_ahead}")
        if not is_tty:
            # We do this because notebooks do not support some escape sequences
            # like the one used to clear the line. So we need to move the cursor
            # to the beginning of the line and overwrite the previous message.
            max_line_length = 73
            return s.ljust(max_line_length, " ")
        return s

    def wait(self,
             polling_period: int = 5,
             download_std_on_completion: bool = True) -> models.TaskStatusCode:
        """Wait for the task to complete.

        This method issues requests to the API.

        Args:
            polling_period: How often to poll the API for the task status.
            download_std_on_completion: Request immediate download of the
                standard files (stdout and stderr) after the task completes.

        Returns:
            The final status of the task.
        """
        # TODO: refactor method to make it cleaner
        prev_status = None
        prev_tasks_ahead = None
        is_tty = sys.stdout.isatty()
        requires_newline = False
        while True:
            status = self.get_status()
            if status != prev_status:
                if requires_newline:
                    requires_newline = False
                    sys.stdout.write("\n")

                if status == models.TaskStatusCode.PENDINGINPUT:
                    pass
                elif status == models.TaskStatusCode.SUBMITTED:
                    logging.info(
                        "■ Task %s successfully queued and waiting to be "
                        "picked-up for execution...", self.id)
                elif status == models.TaskStatusCode.STARTED:
                    logging.info(
                        "■ Task %s has started and is now running "
                        "remotely.", self.id)
                elif status == models.TaskStatusCode.SUCCESS:
                    logging.info("■ Task %s completed successfully.", self.id)
                elif status == models.TaskStatusCode.FAILED:
                    logging.info("■ Task %s failed.", self.id)
                elif status == models.TaskStatusCode.PENDINGKILL:
                    logging.info("■ Task %s is being killed.", self.id)
                elif status == models.TaskStatusCode.KILLED:
                    logging.info("■ Task %s killed.", self.id)
                elif status == models.TaskStatusCode.ZOMBIE:
                    logging.info("■ The machine was terminated while the task "
                                 "was pending.")
                elif status == models.TaskStatusCode.EXECUTERFAILED:
                    info = self.get_info()
                    detail = info.executer.error_detail
                    logging.info(
                        "■ The remote process running the task failed:")
                    if detail:
                        logging.info("\t· Message: %s", detail)
                    else:
                        logging.info("\t· No error message available.")
                elif status == models.TaskStatusCode.SPOTINSTANCEPREEMPTED:
                    msg = ("■ The task was preempted by the cloud provider.\n"
                           "Consider using non-spot machines by setting "
                           "`spot=False` when instantiating the machine group.")
                    logging.info(msg)

                else:
                    logging.info(
                        "■ An internal error occurred with status %s "
                        "while performing the task.", status)
            prev_status = status
            if (status == models.TaskStatusCode.SUBMITTED and
                    self._tasks_ahead is not None and
                    self._tasks_ahead != prev_tasks_ahead):
                requires_newline = True
                sys.stdout.write("\r\033[2K")
                sys.stdout.write(self._setup_queue_message(is_tty))
                sys.stdout.flush()
                prev_tasks_ahead = self._tasks_ahead

            if self.is_terminal():
                sys.stdout.flush()
                sys.stdout.write("\r\033[2K")

                if download_std_on_completion:
                    self._called_from_wait = True
                    out_dir = self.download_outputs(
                        filenames=self.STANDARD_OUTPUT_FILES)
                    if status == models.TaskStatusCode.FAILED:
                        logging.error(
                            "Please inspect the stdout.txt and"
                            " stderr.txt files at %s\n"
                            "For more information.", out_dir)
                return status

            time.sleep(polling_period)

    def _send_kill_request(self, max_api_requests: int) -> None:
        """Send a kill request to the API.
        If the api request fails, it will retry until
        max_api_requests is 0 raising a RuntimeError.
        Args:
            max_api_requests (int): maximum number of api requests to send
        """
        while max_api_requests > 0:
            max_api_requests -= 1
            try:
                if self.is_terminal():
                    break

                path_params = self._get_path_params()
                self._api.kill_task(path_params=path_params)
                break
            except exceptions.ApiException as exc:
                if max_api_requests == 0:
                    raise RuntimeError(
                        "Something went wrong while sending"
                        " the kill command. Please try again later.") from exc
                time.sleep(constants.TASK_KILL_RETRY_SLEEP_SEC)

    def _check_if_pending_kill(
            self,
            wait_timeout: Union[float,
                                int]) -> Tuple[bool, models.TaskStatusCode]:
        """Check if the task is in the PENDINGKILL state.
        This method keeps checking the status of the task until it is no longer
        in the PENDINGKILL state or until the timeout is reached.
        Args:
            wait_timeout (int, float): number of seconds to wait for the
            state to leave PENDINGKILL.
        Returns:
            A tuple with a boolean indicating whether the timeout was reached
            and the status of the task.
        """
        success = True
        start = time.time()

        while (status :=
               self.get_status()) == models.TaskStatusCode.PENDINGKILL:
            if (time.time() - start) > wait_timeout:
                success = False
                break

            time.sleep(constants.TASK_KILL_RETRY_SLEEP_SEC)
        return success, status

    def kill(self,
             wait_timeout: Optional[Union[float, int]] = None,
             verbosity_level: int = 2) -> Union[bool, None]:
        """Request a task to be killed.

        This method requests that the current task is remotely killed.
        If `wait_timeout` is None (default), the kill request is sent to the
        backend and the method returns. However, if `wait_timeout` is
        a positive number, the method waits up to `wait_timeout` seconds
        to ensure that the task transitions to the KILLED state.
        Args:
            wait_timeout (int, float): Optional - number of seconds to wait
            for the kill command or None if only the request is to be sent.
            verbosity_level (int): Optional. the verbosity of the logs when the
            task signal is sent and when the task is killed. Verbosity 0
            produces no outputs, 1 produces minimal outputs, and 2 (Default)
            produces extensive outputs.
        Returns:
            - None if `wait_timeout` is None and the kill request was
              successfully sent;
            - True if `wait_timeout`> 0 and the task successfully transitioned
              to the `KILLED` state within `wait_timeout` seconds;
            - False if `wait_timeout` > 0 but the task didn't transition
              to the `KILLED` state within `wait_timeout` seconds;

        """
        if wait_timeout is not None:
            if not isinstance(wait_timeout, (float, int)):
                raise TypeError("Wait timeout must be a number.")
            if wait_timeout <= 0.0:
                raise ValueError("Wait timeout must be a positive number.")

        if verbosity_level not in self.KILL_VERBOSITY_LEVELS:
            raise ValueError(f"Verbosity {verbosity_level} level not allowed. "
                             f"Choose from {self.KILL_VERBOSITY_LEVELS}")

        self._send_kill_request(constants.TASK_KILL_MAX_API_REQUESTS)

        if wait_timeout is None:
            logging.info(
                __("task-kill-request-sent" + f"-{verbosity_level}", self.id))
            return None

        success, status = self._check_if_pending_kill(wait_timeout)

        if status != models.TaskStatusCode.KILLED:
            success = False
            if status == models.TaskStatusCode.PENDINGKILL:
                logging.error(
                    "Unable to ensure that task %s transitioned to the KILLED "
                    "state after %f seconds. The status of the task is %s.",
                    self.id, wait_timeout, status)
            else:
                logging.error(
                    "Task is already in a terminal state and cannot be killed. "
                    "Current task status is %s.", status)

        if success:
            if verbosity_level == 2:
                logging.info("Successfully killed task %s.", self.id)
            elif verbosity_level == 1:
                logging.info("%s killed.")

        return success

    def get_simulator_name(self) -> str:
        # e.g. retrieve openfoam from fvm.openfoam.run_simulation
        return self.info.method_name.split(".")[1]

    def get_storage_path(self) -> str:
        return self.info.storage_path

    def get_output_info(self) -> output_info.TaskOutputInfo:
        """Get information about the output files of the task.

        Returns:
            An instance of the OutputInfo class, which can be used to
            access info about the output archive (number of files, total
            compressed size, total uncompressed size) and information about
            each file (name, size, compressed size). It can also be used to
            print that information in a formatted way.
        """
        api_response = self._api.get_outputs_list(
            path_params=self._get_path_params())

        archive_info = api_response.body

        output_files = [
            output_info.FileInfo(
                name=file_info["name"],
                size=int(file_info["size"]),
                compressed_size=int(file_info["compressed_size"]),
            ) for file_info in archive_info["contents"]
        ]

        return output_info.TaskOutputInfo(
            total_size_bytes=int(archive_info["size"]),
            files=output_files,
        )

    def _contains_only_std_files(self, output_dir: pathlib.Path) -> bool:
        """Check if the output archive contains only stdout and stderr files.

        Returns:
            True if the output archive contains only stdout and stderr files,
            False otherwise.
        """
        output_files = list(output_dir.iterdir())
        return all(
            file.name in self.STANDARD_OUTPUT_FILES for file in output_files)

    def download_outputs(
        self,
        filenames: Optional[List[str]] = None,
        output_dir: Optional[str] = None,
        uncompress: bool = True,
        rm_downloaded_zip_archive: bool = True,
        rm_remote_files: bool = False,
    ) -> pathlib.Path:
        """Download output files of the task.

        Args:
            filenames: List of filenames to download. If None or empty, all
                files are downloaded.
            output_dir: Directory where to download the files. If None, the
                files are downloaded to the default directory. The default is
                {inductiva.get_output_dir()}/{output_dir}/{task_id}.
            uncompress: Whether to uncompress the archive after downloading it.
            rm_downloaded_zip_archive: Whether to remove the archive after
                uncompressing it. If uncompress is False, this argument is
                ignored.
            rm_remote_files: Whether to remove all task files from remote
                storage after the download is complete. Only used if filenames
                is None or empty (i.e., all output files are downloaded).
        """
        self._status = self.get_status()
        try:
            api_response = self._api.get_output_download_url(
                path_params=self._get_path_params(),)
        except exceptions.ApiException as e:
            if not self._called_from_wait:

                if self._status == models.TaskStatusCode.EXECUTERFAILED:
                    logging.info("The remote process running the task failed:")
                    self.get_info()
                    detail = self.info.executer.error_detail
                    if detail:
                        logging.info(" > Message: %s", detail)
                    else:
                        logging.info(" > No error message available.")
                else:
                    # Raise the exception to be handled by the exception handler
                    raise e
            return None
        finally:
            # Reset internal state
            self._called_from_wait = False

        download_url = api_response.body.get("url")

        if download_url is None:
            raise RuntimeError(
                "The API did not return a download URL for the task outputs.")

        logging.debug("\nDownload URL: %s\n", download_url)

        # If the file server (GCP, ICE, etc.) is not available, the Web API
        # returns a fallback URL and returns the following flag as False.
        # In this case, the output donwload will be provided by the Web API
        # itself.
        file_server_available = bool(
            api_response.body.get("file_server_available"))

        if output_dir is None:
            output_dir = self.id

        output_dir_path = files.resolve_output_path(output_dir)

        if (output_dir_path.exists() and
                not self._contains_only_std_files(output_dir_path)):
            warnings.warn("Path already exists, files may be overwritten.")
        output_dir_path.mkdir(parents=True, exist_ok=True)

        download_message = "Downloading simulation outputs to %s..."

        if filenames is self.STANDARD_OUTPUT_FILES:
            download_message = "Downloading stdout and stderr files to %s..."

        if filenames:
            if file_server_available:
                logging.info(download_message, output_dir)
                data.download_partial_outputs(
                    download_url,
                    filenames,
                    output_dir_path,
                )
            else:
                logging.error("Partial download is not available.")

            # If the user requested a partial download, the full download
            # will be skipped.

            logging.info("Partial download completed to %s.", output_dir)
            return output_dir_path

        zip_path = output_dir_path.joinpath("output.zip")
        logging.info(download_message, zip_path)

        if file_server_available:
            pool_manager: urllib3.PoolManager = (
                self._api.api_client.rest_client.pool_manager)
            response = pool_manager.request(
                "GET",
                download_url,
                preload_content=False,
            )
        # If the file server is not available, the download will be provided
        # by the Web API. Therefore the standard method is used.
        else:
            api_response = self._api.download_task_output(
                path_params=self._get_path_params(),
                stream=True,
                skip_deserialization=True,
            )
            response = api_response.response

        # use raw urllib3 response instead of the generated client response, to
        # implement our own download logic (with progress bar, first checking
        # the size of the file, etc.)
        data.download_file(response, zip_path)

        if uncompress:
            logging.info("Uncompressing the outputs to %s...", output_dir)
            data.uncompress_task_outputs(zip_path, output_dir_path)
            if rm_downloaded_zip_archive:
                zip_path.unlink()

        if rm_remote_files:
            self.remove_remote_files()

        if self._status == models.TaskStatusCode.FAILED:
            logging.error(
                "Task %s failed.\n"
                "Please inspect the stdout.txt and stderr.txt files at %s\n"
                "For more information.", self.id, output_dir)

        return output_dir_path

    class _PathParams(TypedDict):
        """Util class for type checking path params."""
        task_id: str

    def _get_path_params(self) -> _PathParams:
        """Get dictionary with the URL path parameters for API calls."""
        return {"task_id": self.id}

    def _get_duration(
        self,
        start_attribute: str,
        metric_attribute: str,
        cached: bool,
    ) -> Optional[float]:
        """Get the duration of a task phase.

        Args:
            start_attribute: The attribute containing the start time.
            metric_attribute: The attribute containing the duration if the task
                has ended.
            cached: Whether to use the cached info or fetch the latest info.

        Returns:
            The duration in seconds, or None if the start or end time is None.
        """
        info: TaskInfo = self.get_info() if not cached else self.info

        # The task has ended and the metric is available
        metric = getattr(info.time_metrics, metric_attribute)
        if metric.value is not None:
            return metric.value

        # The task has ended but the metric is not available
        if self.info.is_terminal:
            return None

        # The task is still running
        start_time = getattr(info, start_attribute)

        # start time may be None if the task was killed before it started
        if start_time is None:
            return None

        # Format the time to datetime type
        start_time = datetime.datetime.fromisoformat(start_time)
        end_time = datetime.datetime.now(datetime.timezone.utc)

        return (end_time - start_time).total_seconds()

    def get_computation_time(self, cached: bool = False) -> Optional[float]:
        """Get the time the computation of the task took to complete.

        Returns:
            The task computation time if the task is already started or in a
            terminal state, or None otherwise.
        """
        return self._get_duration(
            start_attribute="computation_start_time",
            metric_attribute="computation_seconds",
            cached=cached,
        )

    def get_total_time(self, cached: bool = False) -> Optional[float]:
        """Get the total time the task workflow took to complete.

        Returns:
            The task total duration since it was created, or None if the
            metric is not available or can't be computed.
        """
        return self._get_duration(
            start_attribute="create_time",
            metric_attribute="total_seconds",
            cached=cached,
        )

    def get_machine_type(self) -> Optional[str]:
        """Get the machine type used in the task.

        Streamlines the process of obtaining the task info, extracting the
        machine type from the comprehensive task info.

        Returns:
            The machine type, or None if a machine hasn't been assigned yet.
        """
        info: TaskInfo = self.get_info()
        if info.executer is None:
            return None

        machine_provider = info.executer.host_type
        machine_type = info.executer.vm_type.split("/")[-1]

        return machine_provider + "-" + machine_type

    def remove_remote_files(self) -> None:
        """Removes all files associated with the task from remote storage."""
        logging.info(
            "Removing files from remote storage for task %s...",
            self.id,
        )
        try:
            self._api.delete_task_files(path_params=self._get_path_params())
            logging.info("Remote task files removed successfully.")
        except exceptions.ApiException as e:
            logging.error("An error occurred while removing the files:")
            logging.error(" > %s", json.loads(e.body)["detail"])

    def _get_summary(self) -> str:
        """Get a formatted summary of the task. This method caches the
        summary."""
        info: TaskInfo = self.get_info()

        # Update the duration metrics if the task is still running, otherwise
        # the cached values will be used
        info.time_metrics.total_seconds.value = self.get_total_time(cached=True)
        info.time_metrics.computation_seconds.value = \
            self.get_computation_time(cached=True)

        self._summary = str(info)
        return self._summary

    @property
    def summary(self) -> str:
        """It returns cached information about the task summary."""
        if self._summary is None:
            return self._get_summary()
        return self._summary

    def print_summary(self, fhandle=sys.stdout):
        print(self._get_summary(), file=fhandle)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.projects.post import CreateProject
from inductiva.client.paths.projects_name.get import GetProject
from inductiva.client.paths.projects.get import GetUserProjects


class ProjectsApi(
        CreateProject,
        GetProject,
        GetUserProjects,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



"""List the storage contents via CLI."""

import argparse

from inductiva import storage


def listdir(args):
    """List the user's remote storage contents."""
    storage.listdir(args.path, args.max_results, args.order_by, args.sort_order)


def register(parser):
    """Register the list user's remote storage command."""

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List remote storage contents.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva storage list` command provides an overview"
        " of your data on the platform.\n"
        "It lists all items in a specified path, allowing you to "
        "control the maximum number of results,\n"
        "the ordering criteria, and the sorting order.\n")
    subparser.add_argument("path", default="/", type=str, nargs="?")
    subparser.add_argument("-m", "--max-results", default=10, type=int)
    subparser.add_argument("-o",
                           "--order-by",
                           default="creation_time",
                           type=str,
                           choices=["creation_time", "size"],
                           help="Order by creation_time or size.")
    subparser.add_argument("-s",
                           "--sort-order",
                           default="desc",
                           type=str,
                           choices=["desc", "asc"],
                           help="Sorting order (desc or asc).")

    subparser.set_defaults(func=listdir)



"""Methods to interact with the user storage resources."""
from absl import logging

import inductiva
from inductiva.client import exceptions
from inductiva.client.apis.tags import storage_api
from inductiva.utils import format_utils
from typing import Literal


def get_space_used():
    """Returns the occupied storage size in GB."""
    try:
        api = storage_api.StorageApi(inductiva.api.get_client())
        response = api.get_storage_size()
        storage_used = round(float(response.body) / (1024**3), 3)
        print(f"Total user's remote storage in use: {storage_used} GB")
        return storage_used
    except inductiva.client.ApiException as api_exception:
        raise api_exception


def listdir(path="/",
            max_results: int = 10,
            order_by: Literal["size", "creation_time"] = "creation_time",
            sort_order: Literal["asc", "desc"] = "desc"):
    """List and display the contents of the user's storage.
    Args:
        path (str): Storage directory to list. Default is root.
        max_results (int): The maximum number of results to return.
        order_by (str): The field to sort the contents by.
        sort_order (str): Whether to sort the contents in ascending or
        descending order.
    Returns:
        list of dict: A list of dictionaries containing information about
        the size, the name and the creation time of each content that can
        easily be converted to a dataframe.

    This function prints a table with the storage content information:
        Name            Size            Creation Time
        1234            5.68 MiB        29 Sep, 14:12:00
        12345           374.85 KiB      29 Sep, 14:13:10
        1234567         97.59 KiB       29 Sep, 14:13:24
        123             0 B             29 Sep, 14:13:29
        123456          0 B             29 Sep, 14:13:18
    You can use this information to delete the contents you don't need
    anymore and further inspect task outputs and logs using the Task
    class.
    """
    try:
        api = storage_api.StorageApi(inductiva.api.get_client())
        contents = api.list_storage_contents({
            "path": path,
            "max_results": max_results,
            "sort_by": order_by,
            "order": sort_order
        }).body
        all_contents = []
        for content_name, info in contents.items():
            size = info["size_bytes"]
            creation_time = info["creation_time"]
            all_contents.append({
                "content_name": content_name,
                "size": round(float(size), 3),
                "creation_time": creation_time
            })
        print(_print_contents_table(all_contents))
        return all_contents
    except inductiva.client.ApiException as api_exception:
        raise api_exception


def _print_contents_table(contents):
    columns = ["Name", "Size", "Creation Time"]
    rows = []

    for content in contents:
        row = [
            content["content_name"], content["size"], content["creation_time"]
        ]
        rows.append(row)

    formatters = {
        "Creation Time": [format_utils.datetime_formatter],
        "Size": [format_utils.bytes_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(
        rows,
        columns,
        formatters=formatters,
        header_formatters=header_formatters,
    )


def rmdir(path: str, /, confirm: bool = False):
    """Delete paths inside the user's remote storage in the Inductiva platform.

    This function removes the given `path` from the user's remote storage
    in the Inductiva platform. The given path can either be the storage name of
    a task to remove all content pertaining to that task, or `*` to remove all
    content. The function requires an explicit `confirm=True` keyword argument
    to confirm that the path is, indeed, to be removed.

    E.g.:
        #Works - remote path "task_id" exists and user is explicitly
        # sets confirm=True:
        >>> inductiva.storage.rmdir("task_id", confirm=True)
        
        #Works - The user explicitly sets confirm=True, confirming that
        # all contents in his remote storage space are to be removed:
        >>> inductiva.storage.rmdir("*", confirm=True)
        # Fails -  remote path "task_id" exists but the user fails to
        # confirm his intent. The call will fail win a runtime exception:
        >>> inductiva.storage.rmdir("task_id")

    Args:
        path (str): Path relative to the root of the bucket to delete.
            If path="*" then all the contents in the user's remote storage space
            will be removed.
        confirm (bool): Confirmation flag to explicitly ensure the intent
            to remove the given path. The call will fail with a RuntimeException
            if the user does not set `confirm=True`.
    """
    if not confirm:
        if path == "*":
            path = "all contents"
        raise RuntimeError("Please set `confirm=True` to confirm you want to "
                           f"delete {path} from the user's remote storage.")

    if path == "*":
        logging.info("Removing everything from user's remote storage.")
    else:
        logging.info("Removing %s in the user's remote storage.", path)

    try:
        api = storage_api.StorageApi(inductiva.api.get_client())
        api.delete_path({"path": path})
        logging.info("Successfully removed remote path '%s'.", path)
    except exceptions.ApiException as api_exception:
        if api_exception.status == 404:
            raise exceptions.ApiValueError(
                f"Unable to remove path '{path}'. Path does not exist in "
                "user's remote storage.")
        elif api_exception.status == 500:
            raise exceptions.ApiRuntimeError(
                f"Failed to remove remote path '{path}'.")
        raise api_exception



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskMetrics(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        
        class properties:
            
            
            class total_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'total_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class container_image_download_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'container_image_download_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class queue_time_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'queue_time_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class computation_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'computation_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_upload_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_upload_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_download_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_download_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_decompression_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_decompression_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class output_compression_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'output_compression_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class output_upload_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'output_upload_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_zipped_size_bytes(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_zipped_size_bytes':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_size_bytes(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_size_bytes':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class output_total_files(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'output_total_files':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class output_size_bytes(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'output_size_bytes':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class output_zipped_size_bytes(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'output_zipped_size_bytes':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            __annotations__ = {
                "total_seconds": total_seconds,
                "container_image_download_seconds": container_image_download_seconds,
                "queue_time_seconds": queue_time_seconds,
                "computation_seconds": computation_seconds,
                "input_upload_seconds": input_upload_seconds,
                "input_download_seconds": input_download_seconds,
                "input_decompression_seconds": input_decompression_seconds,
                "output_compression_seconds": output_compression_seconds,
                "output_upload_seconds": output_upload_seconds,
                "input_zipped_size_bytes": input_zipped_size_bytes,
                "input_size_bytes": input_size_bytes,
                "output_total_files": output_total_files,
                "output_size_bytes": output_size_bytes,
                "output_zipped_size_bytes": output_zipped_size_bytes,
            }
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["total_seconds"]) -> MetaOapg.properties.total_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["container_image_download_seconds"]) -> MetaOapg.properties.container_image_download_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["queue_time_seconds"]) -> MetaOapg.properties.queue_time_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["computation_seconds"]) -> MetaOapg.properties.computation_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_upload_seconds"]) -> MetaOapg.properties.input_upload_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_download_seconds"]) -> MetaOapg.properties.input_download_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_decompression_seconds"]) -> MetaOapg.properties.input_decompression_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["output_compression_seconds"]) -> MetaOapg.properties.output_compression_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["output_upload_seconds"]) -> MetaOapg.properties.output_upload_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_zipped_size_bytes"]) -> MetaOapg.properties.input_zipped_size_bytes: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_size_bytes"]) -> MetaOapg.properties.input_size_bytes: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["output_total_files"]) -> MetaOapg.properties.output_total_files: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["output_size_bytes"]) -> MetaOapg.properties.output_size_bytes: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["output_zipped_size_bytes"]) -> MetaOapg.properties.output_zipped_size_bytes: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["total_seconds", "container_image_download_seconds", "queue_time_seconds", "computation_seconds", "input_upload_seconds", "input_download_seconds", "input_decompression_seconds", "output_compression_seconds", "output_upload_seconds", "input_zipped_size_bytes", "input_size_bytes", "output_total_files", "output_size_bytes", "output_zipped_size_bytes", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["total_seconds"]) -> typing.Union[MetaOapg.properties.total_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["container_image_download_seconds"]) -> typing.Union[MetaOapg.properties.container_image_download_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["queue_time_seconds"]) -> typing.Union[MetaOapg.properties.queue_time_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["computation_seconds"]) -> typing.Union[MetaOapg.properties.computation_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_upload_seconds"]) -> typing.Union[MetaOapg.properties.input_upload_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_download_seconds"]) -> typing.Union[MetaOapg.properties.input_download_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_decompression_seconds"]) -> typing.Union[MetaOapg.properties.input_decompression_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["output_compression_seconds"]) -> typing.Union[MetaOapg.properties.output_compression_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["output_upload_seconds"]) -> typing.Union[MetaOapg.properties.output_upload_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_zipped_size_bytes"]) -> typing.Union[MetaOapg.properties.input_zipped_size_bytes, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_size_bytes"]) -> typing.Union[MetaOapg.properties.input_size_bytes, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["output_total_files"]) -> typing.Union[MetaOapg.properties.output_total_files, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["output_size_bytes"]) -> typing.Union[MetaOapg.properties.output_size_bytes, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["output_zipped_size_bytes"]) -> typing.Union[MetaOapg.properties.output_zipped_size_bytes, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["total_seconds", "container_image_download_seconds", "queue_time_seconds", "computation_seconds", "input_upload_seconds", "input_download_seconds", "input_decompression_seconds", "output_compression_seconds", "output_upload_seconds", "input_zipped_size_bytes", "input_size_bytes", "output_total_files", "output_size_bytes", "output_zipped_size_bytes", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        total_seconds: typing.Union[MetaOapg.properties.total_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        container_image_download_seconds: typing.Union[MetaOapg.properties.container_image_download_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        queue_time_seconds: typing.Union[MetaOapg.properties.queue_time_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        computation_seconds: typing.Union[MetaOapg.properties.computation_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_upload_seconds: typing.Union[MetaOapg.properties.input_upload_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_download_seconds: typing.Union[MetaOapg.properties.input_download_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_decompression_seconds: typing.Union[MetaOapg.properties.input_decompression_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        output_compression_seconds: typing.Union[MetaOapg.properties.output_compression_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        output_upload_seconds: typing.Union[MetaOapg.properties.output_upload_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_zipped_size_bytes: typing.Union[MetaOapg.properties.input_zipped_size_bytes, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_size_bytes: typing.Union[MetaOapg.properties.input_size_bytes, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        output_total_files: typing.Union[MetaOapg.properties.output_total_files, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        output_size_bytes: typing.Union[MetaOapg.properties.output_size_bytes, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        output_zipped_size_bytes: typing.Union[MetaOapg.properties.output_zipped_size_bytes, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'TaskMetrics':
        return super().__new__(
            cls,
            *_args,
            total_seconds=total_seconds,
            container_image_download_seconds=container_image_download_seconds,
            queue_time_seconds=queue_time_seconds,
            computation_seconds=computation_seconds,
            input_upload_seconds=input_upload_seconds,
            input_download_seconds=input_download_seconds,
            input_decompression_seconds=input_decompression_seconds,
            output_compression_seconds=output_compression_seconds,
            output_upload_seconds=output_upload_seconds,
            input_zipped_size_bytes=input_zipped_size_bytes,
            input_size_bytes=input_size_bytes,
            output_total_files=output_total_files,
            output_size_bytes=output_size_bytes,
            output_zipped_size_bytes=output_zipped_size_bytes,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Providers(schemas.EnumBase, schemas.StrSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    An enumeration.
    """

    class MetaOapg:
        enum_value_to_name = {
            "GCP": "GCP",
            "ICE": "ICE",
            "LOCAL": "LOCAL",
        }

    @schemas.classproperty
    def GCP(cls):
        return cls("GCP")

    @schemas.classproperty
    def ICE(cls):
        return cls("ICE")

    @schemas.classproperty
    def LOCAL(cls):
        return cls("LOCAL")



"""Methods to interact with the user info on the API."""
import json
from typing import Any, Dict

from inductiva import api
from inductiva.client import ApiClient
from inductiva.client.apis.tags.users_api import UsersApi


def _fetch_quotas_from_api() -> Dict[str, Dict[str, Any]]:
    """Get information about a user's quotas.
    """
    api_config = api.get_api_config()

    with ApiClient(api_config) as client:
        api_instance = UsersApi(client)

        resp = api_instance.get_user_quotas(skip_deserialization=True).response
        quotas = json.loads(resp.data.decode("utf-8"))

        return quotas


def get_quotas() -> Dict[str, Dict[str, Any]]:
    """Get the user quotas.

    This function gets a dict with the user quotas.

    Returns:
        Dict with the user quotas.
    """
    quotas = _fetch_quotas_from_api()

    return quotas


def get_info() -> Dict[str, Any]:
    """Get the user information.

    This funtion gets the user information, including the user's name, email,
    username, tier, programs, and total available credits.

    Returns:
        Dict with the user information.
    """

    api_config = api.get_api_config()
    with (ApiClient(api_config)) as client:
        api_instance = UsersApi(client)
        request = api_instance.get_auth_user_info()
    return request.body



"""CLI commands for listing both the active and available machine groups."""

from typing import TextIO
import argparse
import bisect
import sys

from inductiva.resources.machine_types import ProviderType
from inductiva.resources import machines_base
from inductiva.utils import format_utils
from inductiva import resources, _cli


def pretty_print_machines_info(machines_dict):
    """Format and print the information for given machine dict."""
    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]
    print()
    for family, family_details in machines_dict.items():
        print(f"CPU family: {family}")
        final_table = {"Machine Type": [], "Supported vCPUs": [], "Config": []}
        first_line = True
        for machine_type, details in family_details.items():
            # Used to determine if we write the machine type name
            first_time_type = True

            # Don't want to add an empty line before the first machine type
            if not first_line:
                # Add's an empty line between the machine types
                final_table["Machine Type"].append("")
                final_table["Supported vCPUs"].append("")
                final_table["Config"].append("")
            first_line = False

            for config, vcpus in details.items():
                if first_time_type:
                    final_table["Machine Type"].append(machine_type)
                    # The first time we add the machine type, we don't want to
                    # write its name
                    first_time_type = False
                else:
                    # If we have more than one config for the same machine type
                    # we dont want to repeat the machine type name
                    # (ex c3 standard)
                    final_table["Machine Type"].append("")
                str_vcpus = "n/a" if not vcpus["vcpus"] else ", ".join(
                    str(v) for v in vcpus["vcpus"])
                final_table["Supported vCPUs"].append(str_vcpus)
                final_table["Config"].append(config)

        res_table = format_utils.get_tabular_str(
            final_table,
            header_formatters=header_formatters,
            indentation_level=4)
        print(res_table)

    print("Ex:\tc3d-highcpu-4")
    print("\tc3d-highcpu-8-lssd")


def list_machine_types_available(args):
    """List available machine types information per provider and CPU series."""

    provider = args.provider
    machine_family = args.family

    resources_available = resources.machine_types.get_available_machine_types(
        provider, machine_family)

    machines_dict = {}

    for machine in resources_available:
        # Avoid the duplication of adding machine types with spot instances.
        if machine["spot"]:
            continue

        machine_type = machine["machine_type"]
        machine_info = machine_type.split("-")

        family = machine_info[0]
        memory = machine_info[1]
        vcpus = machine_info[2] if len(machine_info) > 2 else None
        config = machine_info[3] if len(machine_info) > 3 else None

        if family not in machines_dict:
            machines_dict[family] = {}
        if memory not in machines_dict[family]:
            machines_dict[family][memory] = {}
        if config not in machines_dict[family][memory]:
            machines_dict[family][memory][config] = {"vcpus": []}

        if vcpus is not None:
            # Sorted insertion of vcpus
            bisect.insort(machines_dict[family][memory][config]["vcpus"],
                          int(vcpus))
    pretty_print_machines_info(machines_dict)


def _machine_group_list_to_str(machine_group_list) -> str:
    """Returns a string representation of a list of machine groups."""
    columns = [
        "Name", "Machine Type", "Elastic", "Type", "# machines",
        "Data Size in GB", "Spot", "Started at (UTC)", "Idle Time",
        "Max Cost ($/hour)"
    ]
    rows = []

    for machine_group in machine_group_list:
        is_elastic = False
        resource_type = machines_base.ResourceType.STANDARD.value
        spot = machine_group.spot if hasattr(machine_group, "spot") else False

        if isinstance(machine_group, resources.ElasticMachineGroup):
            is_elastic = True
        else:
            if isinstance(machine_group, resources.MPICluster):
                resource_type = machines_base.ResourceType.MPI.value
        num_active_machines = machine_group.active_machines_to_str()
        rows.append([
            machine_group.name, machine_group.machine_type, is_elastic,
            resource_type, num_active_machines, machine_group.data_disk_gb,
            spot, machine_group.create_time,
            f"{machine_group.idle_time}/{machine_group.max_idle_time}",
            machine_group.quota_usage.get("max_price_hour")
        ])

    formatters = {
        "Started at (UTC)": [format_utils.datetime_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(rows,
                                        columns,
                                        formatters=formatters,
                                        header_formatters=header_formatters)


def list_machine_groups(_, fout: TextIO = sys.stdout):
    # pylint: disable=line-too-long
    """Lists all active resources info.

    This method queries Google Cloud for active resources
    that belong to the user. It outputs all the information relative to
    each resource as folllows:

    Active Resources:
                                        Name         VM Type    Elastic   # machines    Disk Size in GB       Spot   Started at (UTC)
    api-6359c03d-c4f9-479f-8b11-ba1f8f55a58c   e2-standard-4      False            3                 40      False   10 Oct, 13:40:50
    api-db2046cf-a6fc-4124-926c-1a24329da5ea   e2-standard-4       True          2/4                 40      False   10 Oct, 12:43:03
    """
    # pylint: enable=line-too-long

    machine_group_list = resources.machine_groups.get()
    if len(machine_group_list) != 0:
        print("Active Resources:", file=fout)
        print(_machine_group_list_to_str(machine_group_list), file=fout, end="")
    else:
        print("No active computational resources found.", file=fout, end="")


def register(parser):
    """Register the list resources commands."""

    subparser = parser.add_parser("available",
                                  help="List available machine types.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.add_argument("-p",
                           "--provider",
                           type=str,
                           default="gcp",
                           choices=[t.value for t in ProviderType],
                           help="Filter the available types by provider.")
    subparser.add_argument("-f",
                           "--family",
                           default=None,
                           type=str,
                           help="Filter the available types by CPU series.")

    subparser.description = (
        "The `inductiva available` command provides a utility for listing the\n"
        "available machine types by provider (default: gcp) and CPU series.\n"
        "The list includes a description of the memory types and vCPUs "
        "available.\n\n")

    subparser.set_defaults(func=list_machine_types_available)

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List currently active resources.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    _cli.utils.add_watch_argument(subparser)

    subparser.description = (
        "The `inductiva resources list` command provides a snapshot "
        "of your active computational resources.\nFrom machine type "
        "to start time, it gives you a comprehensive overview of "
        "your resources in one place.")
    subparser.set_defaults(func=list_machine_groups)



"""Functions for running simulations via Inductiva Web API."""
import os

import pathlib
from typing import Any, Optional, Union
import json
import threading

from absl import logging

from inductiva import tasks, types
from inductiva.api import methods
from inductiva.utils import format_utils, files
from inductiva.resources.machine_types import ProviderType

TASK_METADATA_FILENAME = "task_metadata.json"

_metadata_lock = threading.RLock()


def run_simulation(
    api_method_name: str,
    input_dir: pathlib.Path,
    resubmit_on_preemption: bool = False,
    computational_resources: Optional[types.ComputationalResources] = None,
    provider_id: Optional[Union[ProviderType, str]] = ProviderType.GCP,
    storage_dir: Optional[str] = "",
    api_invoker=None,
    extra_metadata=None,
    simulator=None,
    **kwargs: Any,
) -> tasks.Task:
    """Run a simulation via Inductiva Web API."""

    params = {
        "sim_dir": input_dir,
        **kwargs,
    }
    type_annotations = {
        "sim_dir": pathlib.Path,
    }

    if api_invoker is None:
        api_invoker = methods.invoke_async_api

    if provider_id is not None:
        provider_id = ProviderType(provider_id)

    container_image = kwargs.get("container_image", None)

    task_id = api_invoker(api_method_name,
                          params,
                          type_annotations,
                          resubmit_on_preemption=resubmit_on_preemption,
                          resource_pool=computational_resources,
                          container_image=container_image,
                          storage_path_prefix=storage_dir,
                          provider_id=provider_id,
                          simulator=simulator)
    if computational_resources is not None:
        logging.info("■ Task %s submitted to the queue of the %s.", task_id,
                     computational_resources)
    else:
        logging.info("■ Task %s submitted to the default queue.", task_id)

    task = tasks.Task(task_id)
    if not isinstance(task_id, str):
        raise RuntimeError(
            f"Expected result to be a string with task_id, got {type(task_id)}")

    position = task.get_position_in_queue()
    if position is not None:
        logging.info("Number of tasks ahead in the queue: %s", position)
    else:
        logging.info("Task %s does not have queue information.", task_id)

    if not format_utils.getenv_bool("DISABLE_TASK_METADATA_LOGGING", False):
        machine_group_id = None
        if computational_resources is not None:
            machine_group_id = computational_resources.id

        metadata = {
            "api_method_name": api_method_name.split(".")[1],
            "machine_group_id": machine_group_id,
            "storage_dir": storage_dir,
            **kwargs,
        }
        if extra_metadata is not None:
            metadata = {**metadata, **extra_metadata}

        with _metadata_lock:
            _save_metadata({
                **{
                    "task_id": task_id,
                    "input_dir": str(input_dir)
                },
                **metadata
            })
        logging.info(
            "Task %s configurations metadata saved to the tasks metadata file "
            "%s in the current working directory.", task_id,
            TASK_METADATA_FILENAME)

    logging.info(
        "· Consider tracking the status of the task via CLI:"
        "\n\tinductiva tasks list --id %s", task_id)
    logging.info(
        "· Or, tracking the logs of the task via CLI:"
        "\n\tinductiva logs %s", task_id)
    logging.info(
        "· You can also get more information "
        "about the task via the CLI command:"
        "\n\tinductiva tasks info %s", task_id)
    logging.info("")
    return task


def _save_metadata(metadata, mode="a"):
    """Appends metadata to the TASK_METADATA_FILENAME in the cwd."""

    file_path = files.resolve_output_path(TASK_METADATA_FILENAME)
    if not os.path.exists(file_path.parent):
        os.mkdir(file_path.parent)
    with open(file_path, mode, encoding="utf-8") as f:
        json.dump(metadata, f)
        f.write("\n")
    logging.info("Simulation metadata logged to: %s", file_path)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Campaign(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "available_credits",
            "expiry_date",
            "initial_credits",
            "name",
            "enrollment_date",
        }
        
        class properties:
            name = schemas.StrSchema
            enrollment_date = schemas.DateTimeSchema
            expiry_date = schemas.DateTimeSchema
            initial_credits = schemas.NumberSchema
            available_credits = schemas.NumberSchema
            __annotations__ = {
                "name": name,
                "enrollment_date": enrollment_date,
                "expiry_date": expiry_date,
                "initial_credits": initial_credits,
                "available_credits": available_credits,
            }
    
    available_credits: MetaOapg.properties.available_credits
    expiry_date: MetaOapg.properties.expiry_date
    initial_credits: MetaOapg.properties.initial_credits
    name: MetaOapg.properties.name
    enrollment_date: MetaOapg.properties.enrollment_date
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["enrollment_date"]) -> MetaOapg.properties.enrollment_date: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["expiry_date"]) -> MetaOapg.properties.expiry_date: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["initial_credits"]) -> MetaOapg.properties.initial_credits: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["available_credits"]) -> MetaOapg.properties.available_credits: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["name", "enrollment_date", "expiry_date", "initial_credits", "available_credits", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["enrollment_date"]) -> MetaOapg.properties.enrollment_date: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["expiry_date"]) -> MetaOapg.properties.expiry_date: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["initial_credits"]) -> MetaOapg.properties.initial_credits: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["available_credits"]) -> MetaOapg.properties.available_credits: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["name", "enrollment_date", "expiry_date", "initial_credits", "available_credits", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        available_credits: typing.Union[MetaOapg.properties.available_credits, decimal.Decimal, int, float, ],
        expiry_date: typing.Union[MetaOapg.properties.expiry_date, str, datetime, ],
        initial_credits: typing.Union[MetaOapg.properties.initial_credits, decimal.Decimal, int, float, ],
        name: typing.Union[MetaOapg.properties.name, str, ],
        enrollment_date: typing.Union[MetaOapg.properties.enrollment_date, str, datetime, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'Campaign':
        return super().__new__(
            cls,
            *_args,
            available_credits=available_credits,
            expiry_date=expiry_date,
            initial_credits=initial_credits,
            name=name,
            enrollment_date=enrollment_date,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Task(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "method_name",
            "project",
            "task_id",
            "is_terminated",
            "status",
        }
        
        class properties:
            task_id = schemas.StrSchema
        
            @staticmethod
            def status() -> typing.Type['TaskStatusCode']:
                return TaskStatusCode
            method_name = schemas.StrSchema
            project = schemas.StrSchema
            is_terminated = schemas.BoolSchema
            
            
            class storage_path(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'storage_path':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class container_image(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'container_image':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class create_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'create_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class input_submit_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'input_submit_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class start_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'start_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class computation_start_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'computation_start_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class computation_end_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'computation_end_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class end_time(
                schemas.DateTimeBase,
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'end_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class cost(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'cost':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class storage_size(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'storage_size':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class metrics(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            TaskMetrics,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'metrics':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class executer(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Executer,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'executer':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            __annotations__ = {
                "task_id": task_id,
                "status": status,
                "method_name": method_name,
                "project": project,
                "is_terminated": is_terminated,
                "storage_path": storage_path,
                "container_image": container_image,
                "create_time": create_time,
                "input_submit_time": input_submit_time,
                "start_time": start_time,
                "computation_start_time": computation_start_time,
                "computation_end_time": computation_end_time,
                "end_time": end_time,
                "cost": cost,
                "storage_size": storage_size,
                "metrics": metrics,
                "executer": executer,
            }
    
    method_name: MetaOapg.properties.method_name
    project: MetaOapg.properties.project
    task_id: MetaOapg.properties.task_id
    is_terminated: MetaOapg.properties.is_terminated
    status: 'TaskStatusCode'
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["task_id"]) -> MetaOapg.properties.task_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["status"]) -> 'TaskStatusCode': ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["method_name"]) -> MetaOapg.properties.method_name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["project"]) -> MetaOapg.properties.project: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["is_terminated"]) -> MetaOapg.properties.is_terminated: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["storage_path"]) -> MetaOapg.properties.storage_path: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["container_image"]) -> MetaOapg.properties.container_image: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["create_time"]) -> MetaOapg.properties.create_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["input_submit_time"]) -> MetaOapg.properties.input_submit_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["start_time"]) -> MetaOapg.properties.start_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["computation_start_time"]) -> MetaOapg.properties.computation_start_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["computation_end_time"]) -> MetaOapg.properties.computation_end_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["end_time"]) -> MetaOapg.properties.end_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["cost"]) -> MetaOapg.properties.cost: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["storage_size"]) -> MetaOapg.properties.storage_size: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["metrics"]) -> MetaOapg.properties.metrics: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["executer"]) -> MetaOapg.properties.executer: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["task_id", "status", "method_name", "project", "is_terminated", "storage_path", "container_image", "create_time", "input_submit_time", "start_time", "computation_start_time", "computation_end_time", "end_time", "cost", "storage_size", "metrics", "executer", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["task_id"]) -> MetaOapg.properties.task_id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["status"]) -> 'TaskStatusCode': ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["method_name"]) -> MetaOapg.properties.method_name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["project"]) -> MetaOapg.properties.project: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["is_terminated"]) -> MetaOapg.properties.is_terminated: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["storage_path"]) -> typing.Union[MetaOapg.properties.storage_path, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["container_image"]) -> typing.Union[MetaOapg.properties.container_image, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["create_time"]) -> typing.Union[MetaOapg.properties.create_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["input_submit_time"]) -> typing.Union[MetaOapg.properties.input_submit_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["start_time"]) -> typing.Union[MetaOapg.properties.start_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["computation_start_time"]) -> typing.Union[MetaOapg.properties.computation_start_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["computation_end_time"]) -> typing.Union[MetaOapg.properties.computation_end_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["end_time"]) -> typing.Union[MetaOapg.properties.end_time, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["cost"]) -> typing.Union[MetaOapg.properties.cost, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["storage_size"]) -> typing.Union[MetaOapg.properties.storage_size, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["metrics"]) -> typing.Union[MetaOapg.properties.metrics, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["executer"]) -> typing.Union[MetaOapg.properties.executer, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["task_id", "status", "method_name", "project", "is_terminated", "storage_path", "container_image", "create_time", "input_submit_time", "start_time", "computation_start_time", "computation_end_time", "end_time", "cost", "storage_size", "metrics", "executer", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        method_name: typing.Union[MetaOapg.properties.method_name, str, ],
        project: typing.Union[MetaOapg.properties.project, str, ],
        task_id: typing.Union[MetaOapg.properties.task_id, str, ],
        is_terminated: typing.Union[MetaOapg.properties.is_terminated, bool, ],
        status: 'TaskStatusCode',
        storage_path: typing.Union[MetaOapg.properties.storage_path, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        container_image: typing.Union[MetaOapg.properties.container_image, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        create_time: typing.Union[MetaOapg.properties.create_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        input_submit_time: typing.Union[MetaOapg.properties.input_submit_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        start_time: typing.Union[MetaOapg.properties.start_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        computation_start_time: typing.Union[MetaOapg.properties.computation_start_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        computation_end_time: typing.Union[MetaOapg.properties.computation_end_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        end_time: typing.Union[MetaOapg.properties.end_time, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        cost: typing.Union[MetaOapg.properties.cost, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        storage_size: typing.Union[MetaOapg.properties.storage_size, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        metrics: typing.Union[MetaOapg.properties.metrics, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        executer: typing.Union[MetaOapg.properties.executer, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'Task':
        return super().__new__(
            cls,
            *_args,
            method_name=method_name,
            project=project,
            task_id=task_id,
            is_terminated=is_terminated,
            status=status,
            storage_path=storage_path,
            container_image=container_image,
            create_time=create_time,
            input_submit_time=input_submit_time,
            start_time=start_time,
            computation_start_time=computation_start_time,
            computation_end_time=computation_end_time,
            end_time=end_time,
            cost=cost,
            storage_size=storage_size,
            metrics=metrics,
            executer=executer,
            _configuration=_configuration,
            **kwargs,
        )

from inductiva.client.model.executer import Executer
from inductiva.client.model.task_metrics import TaskMetrics
from inductiva.client.model.task_status_code import TaskStatusCode



"""List the user information via CLI."""
from collections import defaultdict
from typing import TextIO
import argparse
import sys

from inductiva import users, _cli
from inductiva.utils import format_utils


def _quotas_to_dict(quotas):
    """Converts a list of quotas to two dictionaries.
    One for global quotas and one for instance quotas.
    """
    table_global = defaultdict(list)
    table_instance = defaultdict(list)
    for _, quota in quotas.items():
        current = quota["in_use"]
        max_allowed = quota["max_allowed"]
        unit = quota["unit"]
        current_str = (f"{current} {unit}" if current is not None else "N/A")
        max_allowed_str = (f"{max_allowed} {unit}"
                           if max_allowed is not None else "N/A")

        if quota["scope"] == "instance":
            table_instance[""].append(quota["label"])
            table_instance["max allowed"].append(max_allowed_str)
        else:
            table_global[""].append(quota["label"])
            table_global["current usage"].append(current_str)
            table_global["max allowed"].append(max_allowed_str)
    return table_global, table_instance


def _print_quotas(_, fout: TextIO = sys.stdout):
    """ Lists the user's quotas.

    Lists all the user's quotas and the quotas left for the user to use.
    """
    table, table_instance = _quotas_to_dict(users.get_quotas())

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    table = format_utils.get_tabular_str(table,
                                         header_formatters=header_formatters)

    table_instance = format_utils.get_tabular_str(
        table_instance, header_formatters=header_formatters)
    print("■ Global User quotas", end="")
    print(table, file=fout)
    print("■ Instance User quotas", end="")
    print(table_instance, file=fout, end="")
    return 0


def _print_credits_summary(user_info, fout: TextIO = sys.stdout):
    """Prints the user's credits information."""
    total_available_credits = user_info["total_available_credits"]
    available_credits = user_info["tier"]["available_credits"]
    currency = user_info["credits_currency"]
    tier = user_info["tier"]["name"]

    print("■ Credits", file=fout)
    print("", file=fout)
    # pylint: disable=consider-using-f-string
    print("  {0:<25s} {1:10.2f} {2}".format(tier + " (tier)", available_credits,
                                            currency),
          file=fout)
    for program in user_info["campaigns"]:
        print("  {0:<25s} {1:10.2f} {2}".format(program["name"] + " (campaign)",
                                                program["available_credits"],
                                                currency),
              file=fout)

    print("  ----------------------------------------", file=fout)
    print("  {0:<25s} {1:10.2f} {2}".format("Total", total_available_credits,
                                            currency),
          file=fout)


def _campaigns_to_dict(campaigns):
    """Converts a list of campaigns to a dictionary."""
    table = defaultdict(list)
    for program in campaigns:
        table["name"].append(program["name"])
        table["enrollment date"].append(program["enrollment_date"])
        table["expiry date"].append(program["expiry_date"])
        table["available credits"].append(program["available_credits"])
        table["initial credits"].append(program["initial_credits"])
    return table


def get_info(_, fout: TextIO = sys.stdout):
    """ Lists the user's credits.

    Lists all the user's credits and the credits left for the user to use.
    """
    user_info = users.get_info()

    tier = user_info["tier"]["name"]
    username = user_info["username"]
    email = user_info["email"]
    name = user_info["name"] or ""

    print(f"Name: {name}", file=fout)
    print(f"Email: {email}", file=fout)
    print(f"Username: {username}", file=fout)

    print("", file=fout)
    print(f"■ Tier: {tier}", file=fout)
    print("", file=fout)

    _print_credits_summary(user_info, fout=fout)

    table = _campaigns_to_dict(user_info["campaigns"])

    emph_formatter = format_utils.get_ansi_formatter()

    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    formatters = {
        "enrollment date": [format_utils.datetime_formatter_ymd_hm,],
        "expiry date": [format_utils.datetime_formatter_ymd_hm,]
    }

    table = format_utils.get_tabular_str(table,
                                         formatters=formatters,
                                         header_formatters=header_formatters)
    print("", file=fout)
    print("■ Active Campaigns", file=fout)
    if not user_info["campaigns"]:
        print("\n Not currently enrolled in any campaign.\n", file=fout)
    else:
        print(table, file=fout)

    _print_quotas(_, fout)

    print("", file=fout)
    return 0


def register(parser):
    """Register the user credits command."""

    subparser = parser.add_parser("info",
                                  help="List the user's information.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva user info` command provides "
        "an overview of your tier, campaigns and credits.\n"
        "It lists all your campaigns as well as the"
        "credits left for you to use.\n")

    _cli.utils.add_watch_argument(subparser)

    subparser.set_defaults(func=get_info)
